// src/index.ts
import { concatAST, Kind as Kind2 } from "graphql";
import a from "indefinite";

// src/visitor.ts
import {
  TypeInfo,
  visitWithTypeInfo,
  visit,
  Kind
} from "graphql";
function isTransformedField(node) {
  return "parentTypename" in node && "fieldName" in node && "kind" in node;
}
function isTransformedFragment(node) {
  return "kind" in node && node.kind === Kind.FRAGMENT_DEFINITION && "selections" in node;
}
function isTransformedOperation(node) {
  return "kind" in node && node.kind === Kind.OPERATION_DEFINITION && "name" in node && typeof node.name === "string";
}
function hasListType(typeNode) {
  if (typeNode.kind === "ListType") {
    return true;
  }
  if (typeNode.kind === "NamedType") {
    return false;
  }
  return hasListType(typeNode.type);
}
function visitor(ast, schema) {
  const typeInfo = new TypeInfo(schema);
  return visit(ast, visitWithTypeInfo(typeInfo, {
    FragmentSpread: {
      leave(node) {
        const transformedFragmentSpread = {
          kind: node.kind,
          name: node.name
        };
        return transformedFragmentSpread;
      }
    },
    FragmentDefinition: {
      leave(node) {
        const transformedFragment = {
          name: node.name,
          kind: node.kind,
          selections: node.selectionSet.selections
        };
        return transformedFragment;
      }
    },
    OperationDefinition: {
      leave(node) {
        const transformedOperation = {
          kind: node.kind,
          operation: node.operation,
          name: node.name,
          selectionSet: node.selectionSet
        };
        return transformedOperation;
      }
    },
    SelectionSet: {
      leave(node) {
        const transformedSelectionSet = {
          kind: node.kind,
          selections: node.selections
        };
        return transformedSelectionSet;
      }
    },
    Field: {
      leave(node) {
        var _a;
        const parentTypename = typeInfo.getParentType().name;
        const transformed = {
          parentTypename,
          fieldName: node.name,
          kind: hasListType(typeInfo.getFieldDef().astNode.type) ? "connection" : "primitive",
          selections: (_a = node.selectionSet) == null ? void 0 : _a.selections
        };
        return transformed;
      }
    },
    Name: {
      leave(node) {
        return node.value;
      }
    }
  }));
}

// src/index.ts
var APOLLO_IMPORTS = [
  "import { ApolloError } from '@apollo/client';",
  "import { MockedResponse } from '@apollo/client/testing';"
];
var MOCK_FN_INTERFACE = `
  interface MockFn<Variables, Query> {
    ({
      variables,
      result,
      error,
    }: {
      variables?: Variables;
      result?: Query;
      error?: ApolloError;
    }): MockedResponse<Query>;
  }  
  `;
function mockFnTemplate({
  functionName,
  queryVarsType,
  queryResultType,
  documentNode,
  queryResult,
  queryResultVar,
  mocks
}) {
  return `
  export const ${functionName}: MockFn<
  ${queryVarsType},
  ${queryResultType}
> = ({ result, variables, error }) => {
  ${mocks}

  ${queryResult}

  return {
    request: {
      query: ${documentNode},
      variables,
    },
    result: {
      data: result != null ? result : ${queryResultVar},
      error,
    },
  };
};
  `;
}
function mockFnName(type, prefix) {
  if (prefix) {
    return `${prefix}${type}`;
  }
  return a(type, { articleOnly: true }) + type;
}
function mockInstanceName(type) {
  return type[0].toLowerCase() + type.substring(1) + "Mock";
}
function buildResultStr({ selections, addTypename }) {
  let ret = "";
  for (const node of selections) {
    if (node === selections[0] && addTypename) {
      ret += ` __typename: '${node.parentTypename}',`;
    }
    if (node.selections == null) {
      const mockInstance = mockInstanceName(node.parentTypename);
      ret += ` ${node.fieldName}: ${mockInstance}.${node.fieldName},`;
    } else {
      const fieldsStr = buildResultStr({
        selections: node.selections,
        addTypename
      });
      const embed = node.kind === "connection" ? `[{${fieldsStr}}]` : `{ ${fieldsStr} }`;
      ret += ` ${node.fieldName}: ${embed},`;
    }
  }
  return ret;
}
function collectTypes({
  selections,
  fragmentMap,
  types = /* @__PURE__ */ new Set()
}) {
  for (const field of selections) {
    if (isTransformedField(field)) {
      if (field.selections == null) {
        types.add(field.parentTypename);
      } else {
        collectTypes({ selections: field.selections, fragmentMap, types });
      }
    } else {
      const { selections: selections2 } = fragmentMap.get(field.name);
      collectTypes({ selections: selections2, fragmentMap, types });
    }
  }
  return [...types];
}
function inlineFragments(selectionSet, fragmentMap) {
  if (selectionSet == null) {
    return null;
  }
  const { selections, ...rest } = selectionSet;
  const next = { ...rest };
  if (selections || next.kind === Kind2.FRAGMENT_SPREAD) {
    const fragment = fragmentMap.get(next.name);
    const nextSelections = selections || fragment.selections;
    next["selections"] = nextSelections.map((node) => inlineFragments(node, fragmentMap));
  }
  return next;
}
function liftFragmentSelections(selectionSet) {
  if (selectionSet == null) {
    return null;
  }
  const { selections, ...rest } = selectionSet;
  const next = { ...rest };
  if (selections) {
    const nextSelections = [];
    for (const node of selections) {
      if (node.kind === Kind2.FRAGMENT_SPREAD) {
        nextSelections.push(...node.selections);
      } else {
        nextSelections.push(node);
      }
    }
    next["selections"] = nextSelections.map((node) => liftFragmentSelections(node));
  }
  return next;
}
function handleOperationResult(node, fragmentMap, { prefix, addTypename }) {
  const operation = node.operation[0].toUpperCase() + node.operation.slice(1);
  const functionName = `mock${node.name}`;
  const queryResultType = `${node.name}${operation}`;
  const queryVarsType = `${node.name}${operation}Variables`;
  const documentNode = `${node.name}Document`;
  const inlinedFragments = inlineFragments(node.selectionSet, fragmentMap);
  const { selections } = liftFragmentSelections(inlinedFragments);
  const types = collectTypes({ selections, fragmentMap });
  const resultFields = buildResultStr({
    selections,
    addTypename
  });
  const mocks = types.map((typeStr) => {
    const mockFn = mockFnName(typeStr, prefix);
    const mockInstance = mockInstanceName(typeStr);
    return `const ${mockInstance} = ${mockFn}();`;
  }).join("\n");
  const queryResultVar = `${node.name}Result`;
  const queryResult = `const ${queryResultVar}: ${queryResultType} = {${resultFields}}`;
  const templateVars = {
    functionName,
    queryResultType,
    queryVarsType,
    mocks,
    queryResult,
    documentNode,
    queryResultVar
  };
  const imports = [documentNode, queryResultType, queryVarsType];
  return {
    templateVars,
    imports
  };
}
function getConfig(config) {
  const baseConfig = {
    typesFile: "./types.ts",
    addTypename: false
  };
  const resolved = Object.assign(baseConfig, config);
  resolved.typesFile = resolved.typesFile.replace(/\.[\w]+$/, "");
  return resolved;
}
var plugin = (schema, documents, config) => {
  const docsAST = concatAST(documents.map((v) => v.document));
  const result = visitor(docsAST, schema);
  const resolvedConfig = getConfig(config);
  const operations = [];
  const fragmentMap = /* @__PURE__ */ new Map();
  for (const node of result.definitions) {
    if (isTransformedOperation(node)) {
      operations.push(node);
    }
    if (isTransformedFragment(node)) {
      fragmentMap.set(node.name, node);
    }
  }
  const templateVars = operations.map((operation) => handleOperationResult(operation, fragmentMap, resolvedConfig));
  const typeImports = templateVars.flatMap(({ imports: imports2 }) => imports2);
  const imports = [
    ...APOLLO_IMPORTS,
    `import { ${typeImports.join(", ")} } from '${resolvedConfig.typesFile}'`
  ];
  const lines = [MOCK_FN_INTERFACE];
  return {
    prepend: imports,
    content: lines.concat(templateVars.map(({ templateVars: templateVars2 }) => mockFnTemplate(templateVars2))).join("\n")
  };
};
export {
  getConfig,
  plugin
};
